import pandas as pd
import numpy as np
from lhcsmapi.metadata.SignalMetadata import SignalMetadata

from lhcsmapi.metadata.MappingMetadata import MappingMetadata

from pyspark.sql.types import StructType, StructField
from pyspark.sql.types import DoubleType, IntegerType


class BusbarResistanceAnalysis(object):

    @staticmethod
    def fill_missing_rows(results_df, names, prefix, index):
        results_index_df = results_df[results_df[f'{prefix}_class'] == index]
        names_missing = set(names) - set(results_index_df[f'{prefix}_name'])
        if len(names_missing) > 0:
            df_missing = pd.DataFrame(list(names_missing), columns=[f'{prefix}_name'])
            df_missing[f'{prefix}_class'] = index
            return results_index_df.append(df_missing, ignore_index=True, sort=False)
        else:
            return results_index_df

    @staticmethod
    def convert_into_row(u_res_i_meas_results_df, U_RES_names, i_meas_results_df, I_MEAS_devices, query_timing_values,
                         query_timing_names, index):
        values = np.concatenate([u_res_i_meas_results_df['Resistance'].values,
                                 u_res_i_meas_results_df['U_RES_mean_inj'].values,
                                 u_res_i_meas_results_df['U_RES_std_inj'].values,
                                 u_res_i_meas_results_df['U_RES_mean_sb'].values,
                                 u_res_i_meas_results_df['U_RES_std_sb'].values,
                                 i_meas_results_df['I_MEAS_mean_inj'].values,
                                 i_meas_results_df['I_MEAS_std_inj'].values,
                                 i_meas_results_df['I_MEAS_mean_sb'].values,
                                 i_meas_results_df['I_MEAS_std_sb'].values,
                                 query_timing_values])

        R_RES = [name.replace('U_RES', 'R_RES') for name in U_RES_names]
        U_RES_mean_inj = [f'{name}_mean_inj' for name in U_RES_names]
        U_RES_std_inj = [f'{name}_std_inj' for name in U_RES_names]
        U_RES_mean_sb = [f'{name}_mean_sb' for name in U_RES_names]
        U_RES_std_sb = [f'{name}_std_sb' for name in U_RES_names]

        I_MEAS_mean_inj = [f'{name}:I_MEAS_mean_inj' for name in I_MEAS_devices]
        I_MEAS_std_inj = [f'{name}:I_MEAS_std_inj' for name in I_MEAS_devices]
        I_MEAS_mean_sb = [f'{name}:I_MEAS_mean_sb' for name in I_MEAS_devices]
        I_MEAS_std_sb = [f'{name}:I_MEAS_std_sb' for name in I_MEAS_devices]

        COLUMN_NAMES = R_RES + U_RES_mean_inj + U_RES_std_inj + U_RES_mean_sb + U_RES_std_sb + \
                       I_MEAS_mean_inj + I_MEAS_std_inj + I_MEAS_mean_sb + I_MEAS_std_sb + query_timing_names

        return pd.DataFrame(data=[values], columns=COLUMN_NAMES, index=[index])

    @staticmethod
    def calculate_signal_diff(signal_sb_inj_df, circuit_type, signal, system, rows):
        # Extract statistics for injection
        signal_inj_df = BusbarResistanceAnalysis.fill_missing_rows(signal_sb_inj_df, rows, signal, 1)
        # Extract statistics for stable-beams
        signal_sb_df = BusbarResistanceAnalysis.fill_missing_rows(signal_sb_inj_df, rows, signal, 2)
        # Merge current characteristics for injection and stable-beam
        signal_df = BusbarResistanceAnalysis.merge_inj_and_sb(signal_inj_df, signal_sb_df, signal)
        # Append sector name
        signal_df['sector'] = signal_df.apply(
            lambda row: SignalMetadata.get_sector_names_from_components(circuit_type, system,
                                                                        row['{}_name'.format(signal)]), axis=1)
        # Calculate absolute value of current difference
        signal_df['{}_abs_diff'.format(signal)] = signal_df.apply(
            lambda row: abs(row['{}_mean_sb'.format(signal)] - row['{}_mean_inj'.format(signal)]), axis=1)

        return signal_df

    @staticmethod
    def merge_inj_and_sb(results_inj_df, results_sb_df, prefix):
        results_df = results_inj_df.merge(results_sb_df[[f'{prefix}_name', f'{prefix}_mean', f'{prefix}_std']],
                                          on=[f'{prefix}_name'])
        results_df = results_df.sort_values(by=[f'{prefix}_name'])
        # update column SIGNAL_NAMES
        old_cols = [f'{prefix}_mean_x', f'{prefix}_std_x', f'{prefix}_mean_y', f'{prefix}_std_y']
        new_cols = [f'{prefix}_mean_inj', f'{prefix}_std_inj', f'{prefix}_mean_sb', f'{prefix}_std_sb']
        results_df.rename(columns=dict(zip(old_cols, new_cols)), inplace=True)
        # drop class columns
        results_df.drop(columns=f'{prefix}_class', inplace=True)
        return results_df

    @staticmethod
    def sort_according_to_tunnel_pos(circuit_type, u_res_i_meas_results_df):
        u_res_i_meas_results_df['Bus Bar Segment Name'] = u_res_i_meas_results_df.apply(
            lambda row: row['U_RES_name'].split(':')[0], axis=1)
        BUSBAR_NAME_ORDERED = MappingMetadata.get_busbar_to_magnet_mapping(circuit_type)
        u_res_i_meas_results_df['Bus Bar Segment Name'] = pd.Categorical(
            u_res_i_meas_results_df['Bus Bar Segment Name'], categories=BUSBAR_NAME_ORDERED['Bus Bar Segment Name'],
            ordered=True)
        u_res_i_meas_results_df.sort_values(by='Bus Bar Segment Name', inplace=True)
        u_res_i_meas_results_df.reset_index(inplace=True, drop=True)

        return u_res_i_meas_results_df

    @staticmethod
    def concat_timestamp_row(df, timestamp):
        values = df.values[0].tolist()
        values.insert(0, float(timestamp))

        return [values]

    @staticmethod
    def create_schema(df):
        struct_fields = [StructField("timestamp", DoubleType())]
        for col in df.columns:
            struct_fields.append(StructField(col, DoubleType()))

        return StructType(struct_fields)