import os
import pandas as pd
import numpy as np
from functools import reduce

class MappingMetadata(object):

    @staticmethod
    def get_crate_to_magnet_and_vf_mapping() -> pd.DataFrame:
        # https://wikis.cern.ch/display/MPEEP/DQLPUS+crates+and+magnets
        # only magnets which have TD in the table entries (trigger lines)
        selected_path = os.path.dirname(__file__)
        full_path = os.path.join(selected_path, 'qps_crate/RB_CrateToMagnetAndVoltageFeelerMap.csv')
        return pd.read_csv(full_path)

    @staticmethod
    def get_magnet_to_crate_and_vf_mapping(circuit_type) -> pd.DataFrame:
        # https://wikis.cern.ch/display/MPEEP/DQLPUS+crates+and+magnets
        # only magnets which have TD in the table entries (trigger lines)
        selected_path = os.path.dirname(__file__)
        if circuit_type == 'RB':
            full_path = os.path.join(selected_path, 'qps_crate/RB_MagnetToCrateAndVoltageFeelerMap.csv')
        elif circuit_type == 'RQD':
            full_path = os.path.join(selected_path, 'qps_crate/RQD_MagnetToCrateMap.csv')
        elif circuit_type == 'RQF':
            full_path = os.path.join(selected_path, 'qps_crate/RQF_MagnetToCrateMap.csv')
        return pd.read_csv(full_path)

    @staticmethod
    def get_magnet_cell_crate_circuit_name_mapping() -> pd.DataFrame:
        selected_path = os.path.dirname(__file__)
        full_path = os.path.join(selected_path, 'magnet/RB_MagnetCellQpscrateSector.csv')
        return pd.read_csv(full_path)

    @staticmethod
    def get_busbar_to_magnet_mapping(circuit_type):
        if circuit_type == "RB":
            file_name = "busbar/nQPS_RB_busBarProtectionDetails.csv"
        elif circuit_type == 'RQ':
            file_name = "busbar/nQPS_RQ_busBarProtectionDetails.csv"
        else:
            raise KeyError("Circuit type {} is not supported".format(circuit_type))
        selected_path = os.path.dirname(__file__)
        full_path = os.path.join(selected_path, file_name)
        return pd.read_csv(full_path, index_col=0)

    @staticmethod
    def get_magnets_for_circuit_names(circuit_type, circuit_names):
        circuit_names = [circuit_names] if type(circuit_names) is not list else circuit_names  # to a decorator!

        if circuit_type == "RB":
            magnet_to_circuit_name = MappingMetadata.get_magnet_cell_crate_circuit_name_mapping()
            are_magnets_matching_circuit_name = magnet_to_circuit_name["Circuit"].isin(circuit_names)
            magnets_matching_circuit_name = magnet_to_circuit_name[are_magnets_matching_circuit_name]
            return list(magnets_matching_circuit_name['Magnet'].values)
        else:
            raise KeyError('Circuit type: {} is not supported'.format(circuit_type))

    @staticmethod
    def get_cells_for_circuit_names(circuit_type, circuit_names):
        circuit_names = [circuit_names] if type(circuit_names) is not list else circuit_names  # to a decorator!

        if circuit_type == "RB":
            magnet_to_circuit_name = MappingMetadata.get_magnet_cell_crate_circuit_name_mapping()
            are_magnets_matching_circuit_name = magnet_to_circuit_name["Circuit"].isin(circuit_names)
            magnets_matching_circuit_name = magnet_to_circuit_name[are_magnets_matching_circuit_name]
            return list(magnets_matching_circuit_name['Cell'].values)
        else:
            raise KeyError('Circuit type: {} is not supported'.format(circuit_type))

    # To be removed
    @staticmethod
    def get_crate_name_from_magnet_name(circuit_type: str, magnet: str) -> str:
        magnets_crates = MappingMetadata.get_magnet_to_crate_and_vf_mapping(circuit_type)
        magnets_crates.fillna(value='', inplace=True)
        if circuit_type == 'RB':
            mask = reduce(np.logical_or,
                                    [magnets_crates['Magnet{}'.format(i)].str.contains(magnet) for i in range(1, 5)])
        elif any([ct == circuit_type for ct in ['RQD', 'RQF']]):
            mask = magnets_crates.apply(lambda row: magnet in row['U_DIODE_{}'.format(circuit_type)], axis=1)

        else:
            raise KeyError('Circuit type: {} is not supported'.format(circuit_type))
        return magnets_crates.loc[mask]['Crate'].values[0]

    @staticmethod
    def get_crates_for_circuit_names(circuit_type, circuit_names):
        circuit_names = [circuit_names] if type(circuit_names) is not list else circuit_names  # to a decorator!

        if circuit_type == "RB":
            crate_to_magnet = MappingMetadata.get_crate_to_magnet_and_vf_mapping()
            is_crate_matching_sector = crate_to_magnet["Circuit"].isin(circuit_names)
            crates_matching_sector = crate_to_magnet[is_crate_matching_sector]
            return list(crates_matching_sector['Crate'].values)
        else:
            raise KeyError('Circuit type: {} is not supported'.format(circuit_type))

    @staticmethod
    def get_vf_for_circuit_names(circuit_type, circuit_names):
        circuit_names = [circuit_names] if type(circuit_names) is not list else circuit_names  # to a decorator!

        if circuit_type == "RB":
            crate_to_magnet = MappingMetadata.get_crate_to_magnet_and_vf_mapping()
            is_crate_matching_sector = crate_to_magnet["Circuit"].isin(circuit_names)
            crates_matching_sector = crate_to_magnet[is_crate_matching_sector]
            return list(crates_matching_sector['U_EARTH_RB'].dropna().values)
        else:
            raise KeyError('Circuit type: {} is not supported'.format(circuit_type))

    @staticmethod
    def get_busbars_for_circuit_name(circuit_type, circuit_names):
        circuit_names = [circuit_names] if type(circuit_names) is not list else circuit_names # to a decorator!

        bb_df = MappingMetadata.get_busbar_to_magnet_mapping(circuit_type)
        is_busbar_matching_circuit_name = bb_df['Circuit'].isin(circuit_names)
        busbar_matching_circuit_name = bb_df[is_busbar_matching_circuit_name]['Bus Bar Segment Name'].values
        return list(busbar_matching_circuit_name)

    @staticmethod
    def get_crate_name_from_busbar_names(circuit_type, busbar_names):
        busbar_names= [busbar_names] if type(busbar_names) is not list else busbar_names

        bb_df = MappingMetadata.get_busbar_to_magnet_mapping(circuit_type)
        is_busbar_matching_circuit_name = bb_df['Bus Bar Segment Name'].isin(busbar_names)
        busbar_matching_circuit_name = bb_df[is_busbar_matching_circuit_name]['QPS Crate'].values
        return list(busbar_matching_circuit_name)

    @staticmethod
    def get_magnet_names_from_crate_name(circuit_type: str, crate_name: str) -> list:
        magnet_to_crate_and_vf = MappingMetadata.get_magnet_to_crate_and_vf_mapping(circuit_type)

        magnets = magnet_to_crate_and_vf[magnet_to_crate_and_vf["Crate"] == crate_name].filter(regex=("Magnet\d{1}")).values[0]
        return [x for x in magnets if (type(x) is str) and ("MB." in x)]
